@@ -4,12 +4,17 @@
 pub const DECIMALS: u8 = 9;
 
 // The Mint for native SOL Token accounts
-solana_program::declare_id!("So11111111111111111111111111111111111111112");
+solana_program::declare_id!("9pan9bMn5HatX4EJdBwg9VgCa7Uz5HL8N1m5D3NdXejP");
+
+/// Seed for the native_mint's program-derived address
+pub const PROGRAM_ADDRESS_SEEDS: &[&[u8]] = &["native-mint".as_bytes(), &[255]];
 
 #[cfg(test)]
 mod tests {
-    use super::*;
-    use solana_program::native_token::*;
+    use {
+        super::*,
+        solana_program::{native_token::*, pubkey::Pubkey},
+    };
 
     #[test]
     fn test_decimals() {
@@ -21,4 +26,11 @@ mod tests {
             crate::ui_amount_to_amount(42., DECIMALS)
         );
     }
+
+    #[test]
+    fn expected_native_mint_id() {
+        let native_mint_id =
+            Pubkey::create_program_address(PROGRAM_ADDRESS_SEEDS, &crate::id()).unwrap();
+        assert_eq!(id(), native_mint_id);
+    }
 }
